import numpy as np
cimport numpy as cnp

DTYPE = np.float64
ctypedef cnp.float64_t DTYPE_t

ITYPE = np.int32
ctypedef cnp.int32_t ITYPE_t

CTYPE = np.complex128
ctypedef cnp.complex128_t CTYPE_t

CTYPE = np.int64
ctypedef cnp.int64_t LTYPE_t
