cdef extern from "<complex>" namespace "std" nogil:
    double         abs(double complex x)
    double complex acos(double complex x)
    double complex acosh(double complex x)
    double         arg(double complex x)
    double complex asin(double complex x)
    double complex asinh(double complex x)
    double complex atan(double complex x)
    double complex atanh(double complex x)
    double complex conj(double complex x)
    double complex cos(double complex x)
    double complex cosh(double complex x)
    double complex exp(double complex x)
    double         imag(double complex x)
    double complex log(double complex x)
    double complex log10(double complex x)
    double         norm(double complex x)
    double complex proj(double complex x)
    double         real(double complex x)
    double complex sin(double complex x)
    double complex sinh(double complex x)
    double complex sqrt(double complex x)
    double complex tan(double complex x)
    double complex tanh(double complex x)
